@@ -75,6 +75,7 @@ struct _GnomeCupsPrinterDetails {
 typedef struct {
 	guint id;
 	GnomeCupsPrinterAddedCallback func;
+	GnomeCupsOnlyOnceCallback once;
 	gpointer user_data;
 } AddNotify;
 
@@ -129,11 +130,15 @@ gnome_cups_get_printers (void)
 void
 gnome_cups_printer_list_free (GList *printers)
 {
-	GList *l;
-	for (l = printers; l != NULL; l = l->next) {
+	GList *l, *n;
+	
+	l = printers;
+	while (l != NULL) {
+		n = l->next;
 		g_free (l->data);
+		g_list_free_1 (l);
+		l = n;
 	}
-	g_list_free (printers);
 }
 
 char *
@@ -745,12 +750,33 @@ update_printers (void)
 }
 
 static gboolean
-update_printers_timeout (gpointer user_data)
+update_printers_timeout (void)
 {
+	AddNotify *notify;
+	GList *l, *n;
+	
 	/* To avoid unneccessary calls during authentication, check
 	 * if a request is currently executing */
 	if (_gnome_cups_outstanding_request_count () == 0) {
 		update_printers ();
+		
+		/* call all our only_once callbacks and remove them from the notify list */
+		l = add_notifies;
+		while (l != NULL) {
+			n = l->next;
+			
+			notify = l->data;
+			if (notify->once) {
+				notify->once (notify->user_data);
+				add_notifies = g_list_remove_link (add_notifies, l);
+				g_list_free_1 (l);
+				g_free (notify);
+			}
+			
+			l = n;
+		}
+		
+		set_timeout ();
 	}
 	
 	return TRUE;
@@ -795,6 +821,30 @@ gnome_cups_printer_new_printer_notify_add (GnomeCupsPrinterAddedCallback cb,
 	return notify->id;
 }
 
+guint
+gnome_cups_printer_new_printer_notify_add_only_once (GnomeCupsPrinterAddedCallback cb,
+						     GnomeCupsOnlyOnceCallback once,
+						     gpointer user_data)
+{
+	guint id = 0;
+	AddNotify *notify;
+	
+	g_return_val_if_fail (cb != NULL, 0);
+	
+	notify = g_new0 (AddNotify, 1);
+	
+	notify->id = ++id;
+	notify->func = cb;
+	notify->once = once;
+	notify->user_data = user_data;
+	
+	add_notifies = g_list_append (add_notifies, notify);
+	
+	set_timeout ();
+	
+	return notify->id;
+}
+
 void
 gnome_cups_printer_new_printer_notify_remove (guint id)
 {
@@ -804,6 +854,8 @@ gnome_cups_printer_new_printer_notify_remove (guint id)
 		if (notify->id == id) {
 			g_free (notify);
 			add_notifies = g_list_remove_link (add_notifies, l);
+			g_list_free_1 (l);
+			break;
 		}
 	}
 
@@ -841,6 +893,8 @@ gnome_cups_printer_printer_removed_notify_remove (guint id)
 		if (notify->id == id) {
 			g_free (notify);
 			removed_notifies = g_list_remove_link (removed_notifies, l);
+			g_list_free_1 (l);
+			break;
 		}
 	}
 
@@ -1112,10 +1166,10 @@ gnome_cups_printer_get_ppd (GnomeCupsPrinter *printer)
 
 	close (fd);
 	ppd = ppdOpenFile (filename);
-	g_free (filename);
 
 	/* This is loaded in to memory now, so we can free it. */
 	unlink (filename);
+	g_free (filename);
 
 	if (!printer->details->ppd_options)
 		printer->details->ppd_options = get_ppd_options (printer, ppd);
@@ -1682,13 +1736,15 @@ gnome_cups_printer_option_free (GnomeCupsPrinterOption *option)
 void
 gnome_cups_printer_option_list_free (GList *options)
 {
-	GList *l;
-
-	for (l = options; l != NULL; l = l->next) {
+	GList *l, *n;
+	
+	l = options;
+	while (l != NULL) {
+		n = l->next;
 		gnome_cups_printer_option_free (l->data);
+		g_list_free_1 (l);
+		l = n;
 	}
-
-	g_list_free (l);
 }
 
 static void
